#include <msp/gl/extensions/arb_draw_instanced.h>
#include <msp/gl/extensions/msp_primitive_restart.h>
#include "batch.h"
#include "bindable.h"
#include "buffer.h"
#include "error.h"
#include "mesh.h"
#include "vertexarray.h"

using namespace std;

namespace {

template<typename T>
void append(vector<Msp::UInt8> &data, T i)
{
	data.insert(data.end(), sizeof(T), 0);
	*(T *)(&data[data.size()-sizeof(T)]) = i;
}

template<typename T, typename U>
U convert(T n)
{
	if(!static_cast<T>(~n))
		return ~0;
	else
		return n;
}

template<typename T, typename U>
void expand(vector<Msp::UInt8> &data)
{
	unsigned count = data.size()/sizeof(T);
	data.resize(count*sizeof(U));
	for(unsigned i=count; i--;)
		*(U *)(&data[i*sizeof(U)]) = convert<T, U>(*(T *)(&data[i*sizeof(T)]));
}

template<typename T, typename U>
void shrink(vector<Msp::UInt8> &data)
{
	unsigned count = data.size()/sizeof(T);
	for(unsigned i=0; i<count; ++i)
		*(U *)(&data[i*sizeof(U)]) = convert<T, U>(*(T *)(&data[i*sizeof(T)]));
	data.resize(count*sizeof(U));
}

}

namespace Msp {
namespace GL {

unsigned Batch::restart_index = 0;

Batch::Batch(PrimitiveType t):
	prim_type(t),
	index_type(UNSIGNED_SHORT),
	gl_index_type(get_gl_type(index_type)),
	max_index(0),
	restart(false)
{ }

Batch::~Batch()
{
}

void Batch::set_index_type(DataType t)
{
	if(t!=UNSIGNED_SHORT && t!=UNSIGNED_INT)
		throw invalid_argument("Batch::set_data_type");
	if(t==UNSIGNED_SHORT && max_index>0xFFFE)
		throw invalid_operation("Batch::set_data_type");

	if(index_type==UNSIGNED_SHORT && t==UNSIGNED_INT)
		expand<UInt16, UInt32>(data);
	else if(index_type==UNSIGNED_INT && t==UNSIGNED_SHORT)
		shrink<UInt32, UInt16>(data);

	index_type = t;
	gl_index_type = get_gl_type(t);
	update_offset();
	dirty = true;
}

Batch &Batch::append(unsigned i)
{
	append_index(i);

	update_offset();
	dirty = true;

	return *this;
}

Batch &Batch::append(const vector<unsigned> &ind)
{
	if(ind.empty())
		return *this;

	data.reserve(data.size()+ind.size()*get_index_size());
	for(vector<unsigned>::const_iterator i=ind.begin(); i!=ind.end(); ++i)
		append_index(*i);

	update_offset();
	dirty = true;

	return *this;
}

bool Batch::can_append(PrimitiveType other_type)
{
	if(other_type!=prim_type)
		return false;
	else if(prim_type==LINE_STRIP || prim_type==LINE_LOOP || prim_type==TRIANGLE_FAN)
		return MSP_primitive_restart;
	else
		return true;
}

Batch &Batch::append(const Batch &other)
{
	if(other.prim_type!=prim_type)
		throw invalid_argument("Batch::append");
	if(prim_type==LINE_STRIP || prim_type==LINE_LOOP || prim_type==TRIANGLE_FAN)
		static Require _req(MSP_primitive_restart);

	if(other.data.empty())
		return *this;

	// TODO allow appending triangles to a triangle strip

	if(prim_type==POINTS || prim_type==LINES || prim_type==TRIANGLES)
		;
	else if(MSP_primitive_restart)
	{
		restart = true;
		if(index_type==UNSIGNED_INT)
			::append<UInt32>(data, 0xFFFFFFFF);
		else
			::append<UInt16>(data, 0xFFFF);
	}
	else if(prim_type==TRIANGLE_STRIP)
	{
		append(get_index(size()-1));
		append(other.get_index(0));
		if(size()&1)
			append(other.get_index(0));
	}

	unsigned count = other.size();
	for(unsigned i=0; i<count; ++i)
		append_index(other.get_index(i));

	update_offset();
	dirty = true;

	return *this;
}

void Batch::append_index(unsigned i)
{
	if(data.empty())
		max_index = i;
	else
		max_index = max(max_index, i);

	if(index_type==UNSIGNED_SHORT && max_index>0xFFFE)
		set_index_type(UNSIGNED_INT);

	if(index_type==UNSIGNED_INT)
		::append<UInt32>(data, i);
	else
		::append<UInt16>(data, i);
}

unsigned Batch::get_index_size() const
{
	return (index_type==UNSIGNED_INT ? sizeof(UInt32) : sizeof(UInt16));
}

unsigned Batch::get_index(unsigned i) const
{
	if(index_type==UNSIGNED_INT)
		return *(UInt32 *)&data[i*sizeof(UInt32)];
	else
		return *(UInt16 *)&data[i*sizeof(UInt16)];
}

void Batch::draw() const
{
	BindRestore _bind_ibuf(get_buffer(), ELEMENT_ARRAY_BUFFER);
	const void *data_ptr = setup_draw();

	glDrawElements(prim_type, size(), gl_index_type, data_ptr);
}

void Batch::draw_instanced(unsigned count) const
{
	static Require req(ARB_draw_instanced);

	BindRestore _bind_ibuf(get_buffer(), ELEMENT_ARRAY_BUFFER);
	const void *data_ptr = setup_draw();

	glDrawElementsInstanced(prim_type, size(), gl_index_type, data_ptr, count);
}

const void *Batch::setup_draw() const
{
	if(!get_buffer())
		throw invalid_operation("Batch::setup_draw");

	if(restart)
	{
		unsigned index = (index_type==UNSIGNED_INT ? 0xFFFFFFFF : 0xFFFF);

		if(index!=restart_index)
			set_restart_index(index);
	}
	else if(restart_index && restart_index<=max_index)
		set_restart_index(0);

	refresh();

	return reinterpret_cast<const void *>(get_offset());
}

void Batch::set_restart_index(unsigned index)
{
	if(index>0)
	{
		if(!restart_index)
			glEnable(GL_PRIMITIVE_RESTART);
		glPrimitiveRestartIndex(index);
	}
	else
		glDisable(GL_PRIMITIVE_RESTART);

	restart_index = index;
}


Batch::Loader::Loader(Batch &b):
	DataFile::ObjectLoader<Batch>(b)
{
	add("indices", &Loader::indices);
}

void Batch::Loader::indices(const vector<unsigned> &ind)
{
	obj.append(ind);
}

} // namespace GL
} // namespace Msp
