#include <cstring>
#include <set>
#include <msp/core/algorithm.h>
#include <msp/core/hash.h>
#include <msp/core/maputils.h>
#include <msp/gl/extensions/arb_es2_compatibility.h>
#include <msp/gl/extensions/arb_fragment_shader.h>
#include <msp/gl/extensions/arb_gl_spirv.h>
#include <msp/gl/extensions/arb_geometry_shader4.h>
#include <msp/gl/extensions/arb_shader_objects.h>
#include <msp/gl/extensions/arb_uniform_buffer_object.h>
#include <msp/gl/extensions/arb_vertex_shader.h>
#include <msp/gl/extensions/ext_gpu_shader4.h>
#include <msp/gl/extensions/nv_non_square_matrices.h>
#include <msp/io/print.h>
#include <msp/strings/format.h>
#include "buffer.h"
#include "error.h"
#include "misc.h"
#include "program.h"
#include "resources.h"
#include "shader.h"
#include "glsl/compiler.h"

using namespace std;

namespace Msp {
namespace GL {

Program::Program()
{
	init();
}

Program::Program(const std::string &source)
{
	init();

	GlslModule mod;
	mod.set_source(source);
	add_stages(mod);

	link();
	module = 0;
}

Program::Program(const string &vert, const string &frag)
{
	init();

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wdeprecated-declarations"
	attach_shader_owned(new VertexShader(vert));
	attach_shader_owned(new FragmentShader(frag));
#pragma GCC diagnostic pop
	link();
}

Program::Program(const Module &mod, const map<string, int> &spec_values)
{
	init();
	add_stages(mod, spec_values);
	link();
}

void Program::init()
{
	static Require _req(ARB_shader_objects);

	id = glCreateProgram();
	module = 0;
	linked = false;
}

Program::~Program()
{
	for(vector<unsigned>::iterator i=stage_ids.begin(); i!=stage_ids.end(); ++i)
		glDeleteShader(*i);
	glDeleteProgram(id);
}

void Program::add_stages(const Module &mod, const map<string, int> &spec_values)
{
	if(!stage_ids.empty())
		throw invalid_operation("Program::add_stages");

	switch(mod.get_format())
	{
	case Module::GLSL: return add_glsl_stages(static_cast<const GlslModule &>(mod), spec_values);
	case Module::SPIR_V: return add_spirv_stages(static_cast<const SpirVModule &>(mod), spec_values);
	default: throw invalid_argument("Program::add_stages");
	}
}

unsigned Program::add_stage(GLenum type)
{
	switch(type)
	{
	case GL_VERTEX_SHADER: { static Require _req(ARB_vertex_shader); } break;
	case GL_GEOMETRY_SHADER: { static Require _req(ARB_geometry_shader4); } break;
	case GL_FRAGMENT_SHADER: { static Require _req(ARB_fragment_shader); } break;
	default: throw invalid_argument("Program::add_stage");
	}

	unsigned stage_id = glCreateShader(type);
	stage_ids.push_back(stage_id);
	glAttachShader(id, stage_id);

	return stage_id;
}

void Program::add_glsl_stages(const GlslModule &mod, const map<string, int> &spec_values)
{
	module = &mod;

	SL::Compiler compiler;
	compiler.set_source(mod.get_prepared_source(), "<module>");
	compiler.specialize(spec_values);
	compiler.compile(SL::Compiler::PROGRAM);
#ifdef DEBUG
	string diagnostics = compiler.get_diagnostics();
	if(!diagnostics.empty())
		IO::print("Program diagnostics:\n%s\n", diagnostics);
#endif

	vector<SL::Stage::Type> stages = compiler.get_stages();
	for(vector<SL::Stage::Type>::const_iterator i=stages.begin(); i!=stages.end(); ++i)
	{
		unsigned stage_id = 0;
		switch(*i)
		{
		case SL::Stage::VERTEX: stage_id = add_stage(GL_VERTEX_SHADER); break;
		case SL::Stage::GEOMETRY: stage_id = add_stage(GL_GEOMETRY_SHADER); break;
		case SL::Stage::FRAGMENT: stage_id = add_stage(GL_FRAGMENT_SHADER); break;
		default: throw invalid_operation("Program::add_glsl_stages");
		}

		string stage_src = compiler.get_stage_glsl(*i);
		const char *src_ptr = stage_src.data();
		int src_len = stage_src.size();
		glShaderSource(stage_id, 1, &src_ptr, &src_len);

		if(*i==SL::Stage::VERTEX)
		{
			const map<string, unsigned> &attribs = compiler.get_vertex_attributes();
			for(map<string, unsigned>::const_iterator j=attribs.begin(); j!=attribs.end(); ++j)
				glBindAttribLocation(id, j->second, j->first.c_str());
		}

		if(*i==SL::Stage::FRAGMENT && EXT_gpu_shader4)
		{
			const map<string, unsigned> &frag_outs = compiler.get_fragment_outputs();
			for(map<string, unsigned>::const_iterator j=frag_outs.begin(); j!=frag_outs.end(); ++j)
				glBindFragDataLocation(id, j->second, j->first.c_str());
		}

		compile_glsl_stage(stage_id);
	}
}

void Program::compile_glsl_stage(unsigned stage_id)
{
	glCompileShader(stage_id);
	bool compiled = get_shader_i(stage_id, GL_COMPILE_STATUS);

	GLsizei info_log_len = get_shader_i(stage_id, GL_INFO_LOG_LENGTH);
	string info_log(info_log_len+1, 0);
	glGetShaderInfoLog(stage_id, info_log_len+1, &info_log_len, &info_log[0]);
	info_log.erase(info_log_len);
	if(module && module->get_format()==Module::GLSL)
		info_log = static_cast<const GlslModule *>(module)->get_source_map().translate_errors(info_log);

	if(!compiled)
		throw compile_error(info_log);
#ifdef DEBUG
	if(!info_log.empty())
		IO::print("Shader compile info log:\n%s", info_log);
#endif
}

void Program::add_spirv_stages(const SpirVModule &mod, const map<string, int> &spec_values)
{
	static Require _req(ARB_gl_spirv);
	static Require _req2(ARB_ES2_compatibility);

	module = &mod;

	const vector<SpirVModule::EntryPoint> &entry_points = mod.get_entry_points();
	std::set<SpirVModule::Stage> stages;
	for(vector<SpirVModule::EntryPoint>::const_iterator i=entry_points.begin(); i!=entry_points.end(); ++i)
	{
		if(stages.count(i->stage))
			throw invalid_argument("Program::add_spirv_stages");

		switch(i->stage)
		{
		case SpirVModule::VERTEX: add_stage(GL_VERTEX_SHADER); break;
		case SpirVModule::GEOMETRY: add_stage(GL_GEOMETRY_SHADER); break;
		case SpirVModule::FRAGMENT: add_stage(GL_FRAGMENT_SHADER); break;
		default: throw invalid_operation("Program::add_spirv_stages");
		}

		stages.insert(i->stage);
	}

	const vector<UInt32> &code = mod.get_code();
	glShaderBinary(stage_ids.size(), &stage_ids[0], GL_SHADER_BINARY_FORMAT_SPIR_V, &code[0], code.size()*4);

	const vector<SpirVModule::SpecConstant> &spec_consts = mod.get_spec_constants();
	vector<unsigned> spec_id_array;
	vector<unsigned> spec_value_array;
	spec_id_array.reserve(spec_consts.size());
	spec_value_array.reserve(spec_consts.size());
	for(vector<SpirVModule::SpecConstant>::const_iterator i=spec_consts.begin(); i!=spec_consts.end(); ++i)
	{
		map<string, int>::const_iterator j = spec_values.find(i->name);
		if(j!=spec_values.end())
		{
			spec_id_array.push_back(i->constant_id);
			spec_value_array.push_back(j->second);
		}
	}

	vector<SpirVModule::EntryPoint>::const_iterator j=entry_points.begin();
	for(vector<unsigned>::const_iterator i=stage_ids.begin(); i!=stage_ids.end(); ++i, ++j)
		glSpecializeShader(*i, j->name.c_str(), spec_id_array.size(), &spec_id_array[0], &spec_value_array[0]);
}

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wdeprecated-declarations"
void Program::attach_shader(Shader &shader)
{
	unsigned shader_id = shader.steal_id();
	if(!shader_id)
		throw invalid_argument("Program::attach_shader");
	stage_ids.push_back(shader_id);
	compile_glsl_stage(shader_id);
}

void Program::attach_shader_owned(Shader *shader)
{
	attach_shader(*shader);
	delete shader;
}

void Program::detach_shader(Shader &)
{
}

const vector<Shader *> &Program::get_attached_shaders() const
{
	static vector<Shader *> dummy;
	return dummy;
}

void Program::bind_attribute(unsigned index, const string &name)
{
	static Require _req(ARB_vertex_shader);
	glBindAttribLocation(id, index, name.c_str());
}

void Program::bind_attribute(VertexAttribute attr, const string &name)
{
	bind_attribute(get_attribute_semantic(attr), name);
}

void Program::bind_fragment_data(unsigned index, const string &name)
{
	static Require _req(EXT_gpu_shader4);
	glBindFragDataLocation(id, index, name.c_str());
}
#pragma GCC diagnostic pop

void Program::link()
{
	if(stage_ids.empty())
		throw invalid_operation("Program::link");

	uniforms.clear();
	uniform_blocks.clear();
	attributes.clear();

	glLinkProgram(id);
	linked = get_program_i(id, GL_LINK_STATUS);

	GLsizei info_log_len = get_program_i(id, GL_INFO_LOG_LENGTH);
	string info_log(info_log_len+1, 0);
	glGetProgramInfoLog(id, info_log_len+1, &info_log_len, &info_log[0]);
	info_log.erase(info_log_len);
	if(module && module->get_format()==Module::GLSL)
		info_log = static_cast<const GlslModule *>(module)->get_source_map().translate_errors(info_log);

	if(!linked)
		throw compile_error(info_log);
#ifdef DEBUG
	if(!info_log.empty())
		IO::print("Program link info log:\n%s", info_log);
#endif

	if(module->get_format()==Module::GLSL)
	{
		query_uniforms();
		query_attributes();
	}
	else if(module->get_format()==Module::SPIR_V)
	{
		collect_uniforms();
		collect_attributes();
	}

	for(vector<UniformInfo>::const_iterator i=uniforms.begin(); i!=uniforms.end(); ++i)
		require_type(i->type);
	for(vector<AttributeInfo>::const_iterator i=attributes.begin(); i!=attributes.end(); ++i)
		require_type(i->type);
}

void Program::query_uniforms()
{
	unsigned count = get_program_i(id, GL_ACTIVE_UNIFORMS);
	uniforms.reserve(count);
	vector<string> uniform_names(count);
	for(unsigned i=0; i<count; ++i)
	{
		char name[128];
		int len = 0;
		int size;
		GLenum type;
		glGetActiveUniform(id, i, sizeof(name), &len, &size, &type, name);
		if(len && strncmp(name, "gl_", 3))
		{
			/* Some implementations report the first element of a uniform array,
			others report just the name of the array itself. */
			if(len>3 && !strcmp(name+len-3, "[0]"))
				name[len-3] = 0;

			uniforms.push_back(UniformInfo());
			UniformInfo &info = uniforms.back();
			info.name = name;
			info.array_size = size;
			info.type = from_gl_type(type);
			uniform_names[i] = name;
		}
	}

	sort(uniforms, &uniform_name_compare);

	if(ARB_uniform_buffer_object)
	{
		vector<UniformInfo *> uniforms_by_index(count);
		for(unsigned i=0; i<count; ++i)
			if(!uniform_names[i].empty())
				// The element is already known to be present
				uniforms_by_index[i] = &*lower_bound(uniforms, uniform_names[i], &name_search<UniformInfo>);
		query_uniform_blocks(uniforms_by_index);
	}

	uniform_blocks.push_back(UniformBlockInfo());
	UniformBlockInfo &default_block = uniform_blocks.back();

	for(vector<UniformInfo>::iterator i=uniforms.begin(); i!=uniforms.end(); ++i)
		if(!i->block)
		{
			i->location = glGetUniformLocation(id, i->name.c_str());
			i->block = &default_block;
			default_block.uniforms.push_back(&*i);
		}

	default_block.layout_hash = compute_layout_hash(default_block.uniforms);

	update_layout_hash();
}

void Program::query_uniform_blocks(const vector<UniformInfo *> &uniforms_by_index)
{
	std::set<unsigned> used_bind_points;
	unsigned count = get_program_i(id, GL_ACTIVE_UNIFORM_BLOCKS);
	// Reserve an extra index for the default block
	uniform_blocks.reserve(count+1);
	for(unsigned i=0; i<count; ++i)
	{
		char name[128];
		int len;
		glGetActiveUniformBlockName(id, i, sizeof(name), &len, name);
		uniform_blocks.push_back(UniformBlockInfo());
		UniformBlockInfo &info = uniform_blocks.back();
		info.name = name;

		int value;
		glGetActiveUniformBlockiv(id, i, GL_UNIFORM_BLOCK_DATA_SIZE, &value);
		info.data_size = value;

		glGetActiveUniformBlockiv(id, i, GL_UNIFORM_BLOCK_ACTIVE_UNIFORMS, &value);
		vector<int> indices(value);
		glGetActiveUniformBlockiv(id, i, GL_UNIFORM_BLOCK_ACTIVE_UNIFORM_INDICES, &indices[0]);
		for(vector<int>::iterator j=indices.begin(); j!=indices.end(); ++j)
		{
			if(!uniforms_by_index[*j])
				throw logic_error("Program::link");
			info.uniforms.push_back(uniforms_by_index[*j]);
			uniforms_by_index[*j]->block = &info;
		}

		vector<unsigned> query_indices(indices.begin(), indices.end());
		vector<int> values(indices.size());
		glGetActiveUniformsiv(id, query_indices.size(), &query_indices[0], GL_UNIFORM_OFFSET, &values[0]);
		for(unsigned j=0; j<indices.size(); ++j)
			uniforms_by_index[indices[j]]->offset = values[j];

		query_indices.clear();
		for(vector<int>::iterator j=indices.begin(); j!=indices.end(); ++j)
			if(uniforms_by_index[*j]->array_size>1)
				query_indices.push_back(*j);
		if(!query_indices.empty())
		{
			glGetActiveUniformsiv(id, query_indices.size(), &query_indices[0], GL_UNIFORM_ARRAY_STRIDE, &values[0]);
			for(unsigned j=0; j<query_indices.size(); ++j)
				uniforms_by_index[query_indices[j]]->array_stride = values[j];
		}

		query_indices.clear();
		for(vector<int>::iterator j=indices.begin(); j!=indices.end(); ++j)
		{
			DataType t = uniforms_by_index[*j]->type;
			if(is_matrix(t))
				query_indices.push_back(*j);
		}
		if(!query_indices.empty())
		{
			glGetActiveUniformsiv(id, query_indices.size(), &query_indices[0], GL_UNIFORM_MATRIX_STRIDE, &values[0]);
			for(unsigned j=0; j<query_indices.size(); ++j)
				uniforms_by_index[query_indices[j]]->matrix_stride = values[j];
		}

		sort(info.uniforms.begin(), info.uniforms.end(), uniform_location_compare);
		info.layout_hash = compute_layout_hash(info.uniforms);
		unsigned n_bindings = BufferRange::get_n_uniform_buffer_bindings();
		info.bind_point = info.layout_hash%n_bindings;
		while(used_bind_points.count(info.bind_point))
			info.bind_point = (info.bind_point+1)%n_bindings;
		glUniformBlockBinding(id, i, info.bind_point);
		used_bind_points.insert(info.bind_point);
	}
}

void Program::query_attributes()
{
	unsigned count = get_program_i(id, GL_ACTIVE_ATTRIBUTES);
	attributes.reserve(count);
	for(unsigned i=0; i<count; ++i)
	{
		char name[128];
		int len = 0;
		int size;
		GLenum type;
		glGetActiveAttrib(id, i, sizeof(name), &len, &size, &type, name);
		if(len && strncmp(name, "gl_", 3))
		{
			if(len>3 && !strcmp(name+len-3, "[0]"))
				name[len-3] = 0;

			attributes.push_back(AttributeInfo());
			AttributeInfo &info = attributes.back();
			info.name = name;
			info.location = glGetAttribLocation(id, name);
			info.array_size = size;
			info.type = from_gl_type(type);
		}
	}
}

void Program::collect_uniforms()
{
	const SpirVModule &mod = static_cast<const SpirVModule &>(*module);

	// Prepare the default block
	uniform_blocks.push_back(UniformBlockInfo());
	vector<vector<string> > block_uniform_names(1);

	const vector<SpirVModule::Variable> &variables = mod.get_variables();
	for(vector<SpirVModule::Variable>::const_iterator i=variables.begin(); i!=variables.end(); ++i)
	{
		if(i->storage==SpirVModule::UNIFORM && i->struct_type)
		{
			uniform_blocks.push_back(UniformBlockInfo());
			UniformBlockInfo &info = uniform_blocks.back();
			info.name = i->struct_type->name;
			info.bind_point = i->binding;
			info.data_size = i->struct_type->size;

			string prefix;
			if(!i->name.empty())
				prefix = i->struct_type->name+".";
			block_uniform_names.push_back(vector<string>());
			collect_block_uniforms(*i->struct_type, prefix, 0, block_uniform_names.back());
		}
		else if(i->storage==SpirVModule::UNIFORM_CONSTANT && i->location>=0)
		{
			block_uniform_names[0].push_back(i->name);
			uniforms.push_back(UniformInfo());
			UniformInfo &info = uniforms.back();
			info.name = i->name;
			info.location = i->location;
			info.array_size = i->array_size;
			info.type = i->type;
		}
	}

	sort(uniforms, &uniform_name_compare);

	for(unsigned i=0; i<uniform_blocks.size(); ++i)
	{
		UniformBlockInfo &block = uniform_blocks[i];
		const vector<string> &names = block_uniform_names[i];
		for(vector<string>::const_iterator j=names.begin(); j!=names.end(); ++j)
		{
			// The element is already known to be present
			UniformInfo &uni = *lower_bound(uniforms, *j, &name_search<UniformInfo>);
			block.uniforms.push_back(&uni);
			uni.block = &block;
		}
		block.layout_hash = compute_layout_hash(block.uniforms);
	}

	update_layout_hash();
}

void Program::collect_block_uniforms(const SpirVModule::Structure &strct, const string &prefix, unsigned base_offset, vector<string> &uniform_names)
{
	for(vector<SpirVModule::StructMember>::const_iterator i=strct.members.begin(); i!=strct.members.end(); ++i)
	{
		unsigned offset = base_offset+i->offset;
		if(i->struct_type)
		{
			if(i->array_size)
			{
				for(unsigned j=0; j<i->array_size; ++j, offset+=i->array_stride)
					collect_block_uniforms(*i->struct_type, format("%s%s[%d].", prefix, i->name, j), offset, uniform_names);
			}
			else
				collect_block_uniforms(*i->struct_type, prefix+i->name+".", offset, uniform_names);
		}
		else
		{
			string name = prefix+i->name;
			uniform_names.push_back(name);
			uniforms.push_back(UniformInfo());
			UniformInfo &info = uniforms.back();
			info.name = name;
			info.offset = offset;
			info.array_size = i->array_size;
			info.array_stride = i->array_stride;
			info.matrix_stride = i->matrix_stride;
			info.type = i->type;
		}
	}
}

void Program::collect_attributes()
{
	const SpirVModule &mod = static_cast<const SpirVModule &>(*module);

	const vector<SpirVModule::EntryPoint> &entry_points = mod.get_entry_points();
	for(vector<SpirVModule::EntryPoint>::const_iterator i=entry_points.begin(); i!=entry_points.end(); ++i)
		if(i->stage==SpirVModule::VERTEX && i->name=="main")
		{
			for(vector<const SpirVModule::Variable *>::const_iterator j=i->globals.begin(); j!=i->globals.end(); ++j)
				if((*j)->storage==SpirVModule::INPUT)
				{
					attributes.push_back(AttributeInfo());
					AttributeInfo &info = attributes.back();
					info.name = (*j)->name;
					info.location = (*j)->location;
					info.array_size = (*j)->array_size;
					info.type = (*j)->type;
				}
		}
}

void Program::update_layout_hash()
{
	string layout_descriptor;
	for(vector<UniformBlockInfo>::const_iterator i=uniform_blocks.begin(); i!=uniform_blocks.end(); ++i)
		layout_descriptor += format("%d:%x\n", i->bind_point, i->layout_hash);
	uniform_layout_hash = hash32(layout_descriptor);
}

Program::LayoutHash Program::compute_layout_hash(const vector<const UniformInfo *> &uniforms)
{
	string layout_descriptor;
	for(vector<const UniformInfo *>::const_iterator i = uniforms.begin(); i!=uniforms.end(); ++i)
		layout_descriptor += format("%d:%s:%x:%d\n", (*i)->location, (*i)->name, (*i)->type, (*i)->array_size);
	return hash32(layout_descriptor);
}

bool Program::uniform_location_compare(const UniformInfo *uni1, const UniformInfo *uni2)
{
	return uni1->location<uni2->location;
}

bool Program::uniform_name_compare(const UniformInfo &uni1, const UniformInfo &uni2)
{
	return uni1.name<uni2.name;
}

template<typename T>
bool Program::name_search(const T &item, const string &name)
{
	return item.name<name;
}

string Program::get_info_log() const
{
	GLsizei len = get_program_i(id, GL_INFO_LOG_LENGTH);
	string log(len+1, 0);
	glGetProgramInfoLog(id, len+1, &len, &log[0]);
	log.erase(len);
	return log;
}

const Program::UniformBlockInfo &Program::get_uniform_block_info(const string &name) const
{
	for(vector<UniformBlockInfo>::const_iterator i=uniform_blocks.begin(); i!=uniform_blocks.end(); ++i)
		if(i->name==name)
			return *i;
	throw key_error(name);
}

const Program::UniformInfo &Program::get_uniform_info(const string &name) const
{
	vector<UniformInfo>::const_iterator i = lower_bound(uniforms, name, &name_search<UniformInfo>);
	if(i==uniforms.end() || i->name!=name)
		throw key_error(name);
	return *i;
}

int Program::get_uniform_location(const string &name) const
{
	if(name[name.size()-1]==']')
		throw invalid_argument("Program::get_uniform_location");

	vector<UniformInfo>::const_iterator i = lower_bound(uniforms, name, &name_search<UniformInfo>);
	return i!=uniforms.end() && i->name==name && i->block->bind_point<0 ? i->location : -1;
}

const Program::AttributeInfo &Program::get_attribute_info(const string &name) const
{
	vector<AttributeInfo>::const_iterator i = lower_bound(attributes, name, &name_search<AttributeInfo>);
	if(i==attributes.end() || i->name!=name)
		throw key_error(name);
	return *i;
}

int Program::get_attribute_location(const string &name) const
{
	if(name[name.size()-1]==']')
		throw invalid_argument("Program::get_attribute_location");

	vector<AttributeInfo>::const_iterator i = lower_bound(attributes, name, &name_search<AttributeInfo>);
	return i!=attributes.end() && i->name==name ? i->location : -1;
}

void Program::bind() const
{
	if(!linked)
		throw invalid_operation("Program::bind");

	if(!set_current(this))
		return;

	glUseProgram(id);
}

void Program::unbind()
{
	if(!set_current(0))
		return;

	glUseProgram(0);
}


Program::UniformInfo::UniformInfo():
	block(0),
	location(-1),
	array_size(0),
	array_stride(0),
	matrix_stride(0),
	type(VOID)
{ }


Program::UniformBlockInfo::UniformBlockInfo():
	data_size(0),
	bind_point(-1),
	layout_hash(0)
{ }


Program::AttributeInfo::AttributeInfo():
	location(-1),
	array_size(0),
	type(VOID)
{ }


Program::Loader::Loader(Program &p, Collection &c):
	DataFile::CollectionObjectLoader<Program>(p, &c)
{
	add("module",          &Loader::module);

	// Deprecated
	add("attribute",       &Loader::attribute);
	add("fragment_shader", &Loader::fragment_shader);
	add("geometry_shader", &Loader::geometry_shader);
	add("vertex_shader",   &Loader::vertex_shader);
}

void Program::Loader::finish()
{
	obj.link();
}

void Program::Loader::module(const string &n)
{
	map<string, int> spec_values;
	SpecializationLoader ldr(spec_values);
	load_sub_with(ldr);
	obj.add_stages(get_collection().get<Module>(n), spec_values);
}

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wdeprecated-declarations"
void Program::Loader::attribute(unsigned i, const string &n)
{
	obj.bind_attribute(i, n);
}

void Program::Loader::fragment_shader(const string &src)
{
	obj.attach_shader_owned(new FragmentShader(src));
}

void Program::Loader::geometry_shader(const string &src)
{
	obj.attach_shader_owned(new GeometryShader(src));
}

void Program::Loader::vertex_shader(const string &src)
{
	obj.attach_shader_owned(new VertexShader(src));
}
#pragma GCC diagnostic pop


DataFile::Loader::ActionMap Program::SpecializationLoader::shared_actions;

Program::SpecializationLoader::SpecializationLoader(map<string, int> &sv):
	spec_values(sv)
{
	set_actions(shared_actions);
}

void Program::SpecializationLoader::init_actions()
{
	add("specialize", &SpecializationLoader::specialize_bool);
	add("specialize", &SpecializationLoader::specialize_int);
}

void Program::SpecializationLoader::specialize_bool(const string &name, bool value)
{
	spec_values[name] = value;
}

void Program::SpecializationLoader::specialize_int(const string &name, int value)
{
	spec_values[name] = value;
}

} // namespace GL
} // namespace Msp
