#include <algorithm>
#include <msp/strings/format.h>
#include <msp/strings/lexicalcast.h>
#include <msp/strings/utils.h>
#include "error.h"
#include "vertexformat.h"

using namespace std;

namespace Msp {
namespace GL {

VertexFormat::VertexFormat():
	count(0)
{ }

VertexFormat::VertexFormat(VertexAttribute a):
	count(1)
{
	attributes[0] = a;
}

VertexFormat VertexFormat::operator,(VertexAttribute a) const
{
	if(count>=MAX_ATTRIBUTES)
		throw invalid_operation("VertexFormat::operator,");

	VertexFormat r = *this;
	r.attributes[r.count++] = a;

	return r;
}

VertexFormat VertexFormat::operator,(unsigned i) const
{
	if(!count)
		throw invalid_operation("VertexFormat::operator,");

	VertexFormat r = *this;
	unsigned char &a = r.attributes[r.count-1];
	a = make_indexed_attribute(static_cast<VertexAttribute>(a), i);

	return r;
}

bool VertexFormat::operator==(const VertexFormat &other) const
{
	if(count!=other.count)
		return false;
	return equal(attributes, attributes+count, other.attributes);
}

unsigned VertexFormat::stride() const
{
	unsigned s = 0;
	for(const unsigned char *i=begin(); i!=end(); ++i)
		s += get_attribute_size(*i);
	return s;
}

int VertexFormat::offset(VertexAttribute attr) const
{
	unsigned sem = get_attribute_semantic(attr);
	unsigned sz = get_attribute_size(attr);
	unsigned offs = 0;
	for(const unsigned char *i=begin(); i!=end(); ++i)
	{
		if(get_attribute_semantic(*i)==sem)
		{
			if(get_attribute_size(*i)>=sz)
				return offs;
			else
				return -1;
		}
		else
			offs += get_attribute_size(*i);
	}

	return -1;
}


VertexAttribute make_indexed_attribute(VertexAttribute attr, unsigned index)
{
	unsigned base = attr;
	if(attr>=TEXCOORD1 && attr<=TEXCOORD4)
	{
		if(index>=4)
			throw out_of_range("make_indexed_attribute");
	}
	else if(attr>=RAW_ATTRIB1 && attr<=RAW_ATTRIB4)
		base &= 7;
	else if(attr<GENERIC1 || attr>GENERIC4)
		throw invalid_argument("make_indexed_attribute");

	if(static_cast<int>((base>>3)+index)>=31)
		throw out_of_range("make_indexed_attribute");

	return static_cast<VertexAttribute>(base+index*8);
}


bool convert_attribute(const string &str, const char *name, int min_size, int max_size, VertexAttribute &attr, VertexAttribute base_attr)
{
	string::const_iterator i = str.begin();
	for(; *name; ++name, ++i)
		if(*i!=*name)
			return false;
	if(i==str.end() || *i<'0'+min_size || *i>'0'+max_size)
		return false;
	VertexAttribute result = static_cast<VertexAttribute>(base_attr+(*i-'0'-min_size));

	if(++i!=str.end())
	{
		if(*i!='_' || ++i==str.end())
			return false;
		try
		{
			result = make_indexed_attribute(result, lexical_cast<unsigned>(string(i, str.end())));
		}
		catch(...)
		{
			// The operator>> will throw a more appropriate exception
			return false;
		}
	}

	attr = result;
	return true;
}

void operator>>(const LexicalConverter &conv, VertexAttribute &a)
{
	const string &str = conv.get();
	if(str=="NORMAL3")
		a = NORMAL3;
	else if(str.size()==12 && !str.compare(0, 5, "COLOR") && str[5]>='3' && str[5]<='4' && !str.compare(6, 6, "_FLOAT"))
		a = static_cast<VertexAttribute>(COLOR3_FLOAT+(str[5]-'3'));
	else if(str=="COLOR4_UBYTE")
		a = COLOR4_UBYTE;
	else if(str=="TANGENT3")
		a = TANGENT3;
	else if(str=="BINORMAL3")
		a = BINORMAL3;
	else if(!convert_attribute(str, "VERTEX", 2, 4, a, VERTEX2) &&
		!convert_attribute(str, "GROUP", 1, 4, a, GROUP1) &&
		!convert_attribute(str, "WEIGHT", 1, 4, a, WEIGHT1) &&
		!convert_attribute(str, "TEXCOORD", 1, 4, a, TEXCOORD1) &&
		!convert_attribute(str, "GENERIC", 1, 4, a, GENERIC1))
		throw lexical_error(format("conversion of '%s' to VertexAttribute", str));
}

} // namespace GL
} // namespace Msp
