#ifndef MSP_GL_SL_RESOLVE_H_
#define MSP_GL_SL_RESOLVE_H_

#include <map>
#include <set>
#include <string>
#include <vector>
#include "visitor.h"

namespace Msp {
namespace GL {
namespace SL {

/** Forms links between nested blocks in the syntax tree. */
class BlockHierarchyResolver: private TraversingVisitor
{
private:
	bool r_any_resolved;

public:
	BlockHierarchyResolver(): r_any_resolved(false) { }

	bool apply(Stage &s) { r_any_resolved = false; s.content.visit(*this); return r_any_resolved; }

private:
	virtual void enter(Block &);
};

/** Resolves types of variables and base types of other types. */
class TypeResolver: private TraversingVisitor
{
private:
	Stage *stage;
	std::map<TypeDeclaration *, TypeDeclaration *> alias_map;
	std::map<TypeDeclaration *, TypeDeclaration *> array_types;
	NodeList<Statement>::iterator type_insert_point;
	InterfaceBlock *iface_block;
	bool r_any_resolved;

public:
	TypeResolver();

	bool apply(Stage &);

private:
	TypeDeclaration *get_or_create_array_type(TypeDeclaration &);
	void resolve_type(TypeDeclaration *&, const std::string &, bool);
	virtual void visit(Block &);
	virtual void visit(BasicTypeDeclaration &);
	virtual void visit(ImageTypeDeclaration &);
	virtual void visit(StructDeclaration &);
	virtual void visit(VariableDeclaration &);
	virtual void visit(InterfaceBlock &);
	virtual void visit(FunctionDeclaration &);
};

/** Resolves variable references.  Variable references which match the name
of an interface block are turned into interface block references. */
class VariableResolver: private TraversingVisitor
{
private:
	Stage *stage;
	RefPtr<Expression> r_replacement_expr;
	bool r_any_resolved;
	bool record_target;
	bool r_self_referencing;
	Assignment::Target r_assignment_target;
	std::vector<VariableDeclaration *> redeclared_builtins;
	std::set<Node *> nodes_to_remove;

public:
	VariableResolver();

	bool apply(Stage &);

private:
	virtual void enter(Block &);
	virtual void visit(RefPtr<Expression> &);
	void check_assignment_target(Statement *);
	virtual void visit(VariableReference &);
	virtual void visit(InterfaceBlockReference &);
	void add_to_chain(Assignment::Target::ChainType, unsigned);
	virtual void visit(MemberAccess &);
	virtual void visit(Swizzle &);
	virtual void visit(BinaryExpression &);
	virtual void visit(Assignment &);
	void merge_layouts(Layout &, const Layout &);
	virtual void visit(VariableDeclaration &);
	virtual void visit(InterfaceBlock &);
};

/** Resolves types and lvalueness of expressions. */
class ExpressionResolver: private TraversingVisitor
{
private:
	enum Compatibility
	{
		NOT_COMPATIBLE,
		LEFT_CONVERTIBLE,
		RIGHT_CONVERTIBLE,
		SAME_TYPE
	};

	struct ArgumentInfo
	{
		BasicTypeDeclaration *type;
		unsigned component_count;
	};

	Stage *stage;
	std::vector<BasicTypeDeclaration *> basic_types;
	NodeList<Statement>::iterator insert_point;
	bool r_any_resolved;

public:
	ExpressionResolver();

	bool apply(Stage &);

private:
	static bool is_scalar(BasicTypeDeclaration &);
	static bool is_vector_or_matrix(BasicTypeDeclaration &);
	static BasicTypeDeclaration *get_element_type(BasicTypeDeclaration &);
	static bool can_convert(BasicTypeDeclaration &, BasicTypeDeclaration &);
	static Compatibility get_compatibility(BasicTypeDeclaration &, BasicTypeDeclaration &);
	BasicTypeDeclaration *find_type(BasicTypeDeclaration::Kind, unsigned);
	BasicTypeDeclaration *find_type(BasicTypeDeclaration &, BasicTypeDeclaration::Kind, unsigned);
	void convert_to(RefPtr<Expression> &, BasicTypeDeclaration &);
	bool convert_to_element(RefPtr<Expression> &, BasicTypeDeclaration &);
	bool truncate_vector(RefPtr<Expression> &, unsigned);
	void resolve(Expression &, TypeDeclaration *, bool);

	virtual void visit(Block &);
	virtual void visit(Literal &);
	virtual void visit(VariableReference &);
	virtual void visit(InterfaceBlockReference &);
	virtual void visit(MemberAccess &);
	virtual void visit(Swizzle &);
	virtual void visit(UnaryExpression &);
	void visit(BinaryExpression &, bool);
	virtual void visit(BinaryExpression &);
	virtual void visit(Assignment &);
	virtual void visit(TernaryExpression &);
	void visit_constructor(FunctionCall &);
	virtual void visit(FunctionCall &);
	virtual void visit(BasicTypeDeclaration &);
	virtual void visit(VariableDeclaration &);
};

/** Resolves function declarations and calls. */
class FunctionResolver: private TraversingVisitor
{
private:
	Stage *stage;
	std::map<std::string, std::vector<FunctionDeclaration *> > declarations;
	bool r_any_resolved;

public:
	bool apply(Stage &);

private:
	virtual void visit(FunctionCall &);
	virtual void visit(FunctionDeclaration &);
};

} // namespace SL
} // namespace GL
} // namespace Msp

#endif
