#ifndef MSP_GL_SL_VALIDATE_H_
#define MSP_GL_SL_VALIDATE_H_

#include <string>
#include <vector>
#include "glsl_error.h"
#include "visitor.h"

namespace Msp {
namespace GL {
namespace SL {

class Validator: protected TraversingVisitor
{
protected:
	Stage *stage;
	Node *last_provoker;

	Validator();

	void diagnose(Node &, Node &, Diagnostic::Severity, const std::string &);
	void diagnose(Node &n, Diagnostic::Severity s, const std::string &m) { diagnose(n, n, s, m); }
	void error(Node &n, const std::string &m) { diagnose(n, Diagnostic::ERR, m); }
	void add_info(Node &, const std::string &);
};

class DeclarationValidator: private Validator
{
private:
	enum ScopeType
	{
		GLOBAL,
		STRUCT,
		INTERFACE_BLOCK,
		FUNCTION_PARAM,
		FUNCTION
	};

	ScopeType scope;
	InterfaceLayout *iface_layout;
	InterfaceBlock *iface_block;
	VariableDeclaration *variable;

public:
	DeclarationValidator();

	void apply(Stage &s) { stage = &s; s.content.visit(*this); }

private:
	static const char *describe_variable(ScopeType);

	virtual void visit(Layout &);
	virtual void visit(InterfaceLayout &);
	virtual void visit(BasicTypeDeclaration &);
	virtual void visit(ImageTypeDeclaration &);
	virtual void visit(StructDeclaration &);
	virtual void visit(VariableDeclaration &);
	virtual void visit(InterfaceBlock &);
	virtual void visit(FunctionDeclaration &);
};

class IdentifierValidator: private Validator
{
private:
	typedef std::map<std::string, Statement *> BlockDeclarationMap;

	std::map<Block *, BlockDeclarationMap> declarations;
	std::map<std::string, InterfaceBlock *> interface_blocks;
	std::map<std::string, FunctionDeclaration *> overloaded_functions;
	bool anonymous_block;

public:
	IdentifierValidator();

	void apply(Stage &s) { stage = &s; s.content.visit(*this); }

private:
	void multiple_definition(const std::string &, Statement &, Statement &);
	Statement *find_definition(const std::string &);
	void check_definition(const std::string &, Statement &);
	void record_definition(const std::string &, Statement &);

	virtual void visit(TypeDeclaration &);
	virtual void visit(BasicTypeDeclaration &t) { visit(static_cast<TypeDeclaration &>(t)); }
	virtual void visit(ImageTypeDeclaration &t) { visit(static_cast<TypeDeclaration &>(t)); }
	virtual void visit(StructDeclaration &);
	virtual void visit(VariableDeclaration &);
	virtual void visit(InterfaceBlock &);
	virtual void visit(FunctionDeclaration &);
};

class ReferenceValidator: private Validator
{
public:
	void apply(Stage &s) { stage = &s; s.content.visit(*this); }

private:
	virtual void visit(BasicTypeDeclaration &);
	virtual void visit(ImageTypeDeclaration &);
	virtual void visit(VariableReference &);
	virtual void visit(MemberAccess &);
	virtual void visit(InterfaceBlockReference &);
	virtual void visit(FunctionCall &);
	virtual void visit(VariableDeclaration &);
	virtual void visit(InterfaceBlock &);
	virtual void visit(FunctionDeclaration &);
};

class ExpressionValidator: private Validator
{
private:
	FunctionDeclaration *current_function;

public:
	ExpressionValidator();

	void apply(Stage &s) { stage = &s; s.content.visit(*this); }

private:
	virtual void visit(Swizzle &);
	virtual void visit(UnaryExpression &);
	virtual void visit(BinaryExpression &);
	virtual void visit(Assignment &);
	virtual void visit(TernaryExpression &);
	virtual void visit(VariableDeclaration &);
	virtual void visit(FunctionDeclaration &);
	virtual void visit(Return &);
};

class StageInterfaceValidator: private Validator
{
private:
	std::map<std::string, std::map<unsigned, VariableDeclaration *> > used_locations;

public:
	void apply(Stage &s) { stage = &s; s.content.visit(*this); }

private:
	int get_location(const Layout &);

	virtual void visit(VariableDeclaration &);
	virtual void visit(FunctionDeclaration &) { }
};

class GlobalInterfaceValidator: private Validator
{
private:
	struct Uniform
	{
		Node *node;
		TypeDeclaration *type;
		std::string name;
		int location;
		unsigned loc_count;
		int desc_set;
		int bind_point;

		Uniform(): node(0), type(0), location(-1), loc_count(1), desc_set(0), bind_point(-1) { }
	};

	std::list<Uniform> uniforms;
	std::map<std::string, const Uniform *> used_names;
	std::map<unsigned, const Uniform *> used_locations;
	std::map<unsigned, std::map<unsigned, const Uniform *> > used_bindings;

public:
	void apply(Module &);

private:
	void check_uniform(const Uniform &);

	virtual void visit(VariableDeclaration &);
	virtual void visit(InterfaceBlock &);
	virtual void visit(FunctionDeclaration &) { }
};

} // namespace SL
} // namespace GL
} // namespace Msp

#endif
